/*
 * Copyright (C) 2015  Roman Pauer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <stdint.h>
#include "hqx-format.h"
#include "hqx-common.h"


#define yuv_diff(yuv1, yuv2) (_cmpgtu4(_subabs4(yuv1, yuv2), 0x00300706))


#if ((HQX_FORMAT_16 == HQX_FORMAT_RGB565) || (HQX_FORMAT_16 == HQX_FORMAT_BGR565))
    #define G_SZ (6)
    #define G_MASK (63 << 2)
#elif ((HQX_FORMAT_16 == HQX_FORMAT_RGB555) || (HQX_FORMAT_16 == HQX_FORMAT_BGR555))
    #define G_SZ (5)
    #define G_MASK (31 << 3)
#else
    #error Bad format selected - HQX_FORMAT_16
#endif

#if ((HQX_FORMAT_16 == HQX_FORMAT_RGB565) || (HQX_FORMAT_16 == HQX_FORMAT_RGB555))
    #define RB_0 r
    #define RB_2 b
#else
    #define RB_0 b
    #define RB_2 r
#endif

#define G_SH (G_SZ - 3)
#define RB_SH_2 (G_SZ + 2)

void prepare_yuv_palette_rgb_16(const uint16_t * restrict RGBpal, uint64_t * restrict RGBYUVpal)
{
    _nassert((int)RGBpal % 8 == 0);
    _nassert((int)RGBYUVpal % 8 == 0);

    unsigned int i;

    for (i = 0; i < 256; i++)
    {
        uint32_t rgbvalue;
        int32_t r, g, b;
        uint32_t Y, u, v;

        rgbvalue = RGBpal[i];

        RB_0 = (rgbvalue << 3      ) & (31 << 3);
        g    = (rgbvalue >> G_SH   ) & G_MASK;
        RB_2 = (rgbvalue >> RB_SH_2) & (31 << 3);

        RB_0 |= RB_0 >> 5;
        g    |= g    >> G_SZ;
        RB_2 |= RB_2 >> 5;

        Y = (r + g + b) >> 2;
        u = 128 + ((r - b) >> 2);
        v = 128 + ((-r + 2*g -b)>>3);

        RGBYUVpal[i] = ((Y << 16) + (u << 8) + v) | (((uint64_t)rgbvalue) << 32);
    }
}

void convert_yuv_8_rgb_16(const uint8_t * restrict src, uint32_t * restrict yuvdst, uint16_t * restrict rgbdst, const uint64_t * restrict RGBYUVpal, int width)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)yuvdst % 8 == 0);
    _nassert((int)rgbdst % 8 == 0);
    _nassert((int)RGBYUVpal % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int index;

    for (index = 0; index < width; index++)
    {
        uint64_t palvalue;

        palvalue = RGBYUVpal[src[index]];

        yuvdst[index] = _loll(palvalue);
        rgbdst[index] = _hill(palvalue);
    }

    yuvdst[-1] = yuvdst[0];
    yuvdst[width] = yuvdst[width - 1];

    rgbdst[-1] = rgbdst[0];
    rgbdst[width] = rgbdst[width - 1];
}

void convert_yuv_16(uint16_t * restrict src, uint32_t * restrict dst, int width)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int index;

    src[-1] = src[0];
    src[width] = src[width - 1];

    for (index = 0; index < width; index++)
    {
        uint32_t rgbvalue;
        int32_t r, g, b;
        uint32_t Y, u, v;

        rgbvalue = src[index];

        RB_0 = (rgbvalue << 3      ) & (31 << 3);
        g    = (rgbvalue >> G_SH   ) & G_MASK;
        RB_2 = (rgbvalue >> RB_SH_2) & (31 << 3);

        RB_0 |= RB_0 >> 5;
        g    |= g    >> G_SZ;
        RB_2 |= RB_2 >> 5;

        Y = (r + g + b) >> 2;
        u = 128 + ((r - b) >> 2);
        v = 128 + ((-r + 2*g -b)>>3);

        dst[index] = (Y << 16) + (u << 8) + v;
    }

    dst[-1] = dst[0];
    dst[width] = dst[width - 1];
}

void calculate_pattern(const uint32_t * restrict yuvsrc1, const uint32_t * restrict yuvsrc2, const uint32_t * restrict yuvsrc3, uint8_t * restrict dst, int width)
{
    _nassert((int)yuvsrc1 % 8 == 0);
    _nassert((int)yuvsrc2 % 8 == 0);
    _nassert((int)yuvsrc3 % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    uint32_t e;
    uint8_t pattern;

    // before first
    {
        e = yuvsrc2[-1];
        pattern = 0;
        if (yuv_diff(e, yuvsrc1[-1])) pattern |= (1 << 0) | (1 << 1) | (1 << 2);
        if (yuv_diff(e, yuvsrc3[-1])) pattern |= (1 << 5) | (1 << 6) | (1 << 7);

        dst[-1] = pattern;
    }

    for (; width != 0; width--)
    {
        e = yuvsrc2[0];
        pattern = 0;
        if (yuv_diff(e, yuvsrc1[-1])) pattern |= 1 << 0;
        if (yuv_diff(e, yuvsrc1[ 0])) pattern |= 1 << 1;
        if (yuv_diff(e, yuvsrc1[ 1])) pattern |= 1 << 2;
        if (yuv_diff(e, yuvsrc2[-1])) pattern |= 1 << 3;
        if (yuv_diff(e, yuvsrc2[ 1])) pattern |= 1 << 4;
        if (yuv_diff(e, yuvsrc3[-1])) pattern |= 1 << 5;
        if (yuv_diff(e, yuvsrc3[ 0])) pattern |= 1 << 6;
        if (yuv_diff(e, yuvsrc3[ 1])) pattern |= 1 << 7;

        *dst = pattern;

        yuvsrc1++;
        yuvsrc2++;
        yuvsrc3++;
        dst++;
    }

    // after last
    {
        e = yuvsrc2[0];
        pattern = 0;
        if (yuv_diff(e, yuvsrc1[ 0])) pattern |= (1 << 0) | (1 << 1) | (1 << 2);
        if (yuv_diff(e, yuvsrc3[ 0])) pattern |= (1 << 5) | (1 << 6) | (1 << 7);

        *dst = pattern;
    }
}

